package com.supwisdom.dlpay.consume

import com.supwisdom.dlpay.consume.domain.TAccount
import com.supwisdom.dlpay.consume.domain.TPerson
import com.supwisdom.dlpay.consume.domain.TUserdtl
import com.supwisdom.dlpay.consume.service.AccountUtilServcie
import com.supwisdom.dlpay.consume.service.PersonBalancePayService
import com.supwisdom.dlpay.exception.TransactionCheckException
import com.supwisdom.dlpay.framework.domain.TShopacc
import com.supwisdom.dlpay.framework.domain.TSubject
import com.supwisdom.dlpay.framework.domain.TTranstype
import com.supwisdom.dlpay.framework.util.TradeErrorCode

class AccountHolder<T> private constructor(val accountId: String, val idType: Int) {
    companion object {
        const val IDTYPE_PERSON = 1
        const val IDTYPE_SHOP = 2
        const val IDTYPE_SUBJECT = 3
        const val IDTYPE_TRANSTYPE = 4

        fun person(userid: String): AccountHolder<TAccount> {
            return AccountHolder<TAccount>(userid, IDTYPE_PERSON)
        }

        fun shop(shopid: String): AccountHolder<TShopacc> {
            return AccountHolder<TShopacc>(shopid, IDTYPE_SHOP)
        }

        fun subject(subjNo: String): AccountHolder<TSubject> {
            return AccountHolder(subjNo, IDTYPE_SUBJECT)
        }

        fun transType(transType: Int): AccountHolder<TTranstype> {
            return AccountHolder("$transType", IDTYPE_TRANSTYPE)
        }
    }

    private lateinit var builder: PersonTransBuilder

    @Suppress("UNCHECKED_CAST", "IMPLICIT_CAST_TO_ANY")
    fun <T> get(): T {
        return when (idType) {
            IDTYPE_PERSON -> builder.accountUtil.readAccount(accountId)
            IDTYPE_SHOP -> builder.accountUtil.readShopAcc(accountId)
            IDTYPE_SUBJECT -> builder.accountUtil.readSubject(accountId)
            IDTYPE_TRANSTYPE -> builder.accountUtil.readTranstype(accountId.toInt())
            else -> throw TransactionCheckException(TradeErrorCode.INPUT_DATA_ERROR, "账户类型未知type<$idType>")
        } as T
    }

    fun withLock(nowait: Boolean): T {
        TODO("not implement")
    }
}

class PersonTransBuilder private constructor(accUitl: AccountUtilServcie) {
    companion object {
        fun newBuilder(accUitl: AccountUtilServcie) = PersonTransBuilder(accUitl)
    }

    inner class TransDetail(val debitAccNo: String, val debitSubjNo: String,
                            val creditAccNo: String, val creditSubjNo: String,
                            val amount: Double, val summary: String,
                            val rowno: Int)

    // 输入参数，调用接口时指定的参数值
    lateinit var person: TPerson

    var transDate = ""
    var transTime = ""
    /**
     * 是否允许透支
     */
    var overdraft = false
    var description = ""

    // 内部参数，不需要调用者处理
    val details = mutableListOf<TransDetail>()
    var amount: Double = 0.0
    lateinit var accountUtil: AccountUtilServcie


    fun setUserAndAcc(per: TPerson) {
        this.person = per
    }

    fun enableOverdraft(b: Boolean = false): PersonTransBuilder {
        this.overdraft = b
        return this
    }

    fun setTransDatetime(date: String, time: String): PersonTransBuilder {
        this.transDate = date
        this.transTime = time
        return this
    }

    private fun prepareData() {
        amount = 0.0
        amount = this.details.sumByDouble { it.amount }
    }

    private fun preCheckAccount() {
    }

    fun <T, U> addDetail(debit: AccountHolder<T>, credit: AccountHolder<U>,
                         amount: Double, summary: String): PersonTransBuilder {
        var debitAccNo = ""
        var debitSubjNo = ""
        when (debit.idType) {
            AccountHolder.IDTYPE_PERSON -> debit.get<TAccount>().also {
                debitAccNo = it.accno
                debitSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_SHOP -> debit.get<TShopacc>().also {
                debitAccNo = it.shopaccno
                debitSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_SUBJECT -> debit.get<TSubject>().also {
                debitAccNo = it.subjno
                debitSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_TRANSTYPE -> TODO("not implements")
        }


        var creditAccNo = ""
        var creditSubjNo = ""
        when (credit.idType) {
            AccountHolder.IDTYPE_PERSON -> credit.get<TAccount>().also {
                creditAccNo = it.accno
                creditSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_SHOP -> credit.get<TShopacc>().also {
                creditAccNo = it.shopaccno
                creditSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_SUBJECT -> credit.get<TSubject>().also {
                creditAccNo = it.subjno
                creditSubjNo = it.subjno
            }
            AccountHolder.IDTYPE_TRANSTYPE -> TODO("not implements")
        }

        this.details.add(TransDetail(debitAccNo, debitSubjNo, creditAccNo, creditSubjNo,
                amount, summary, this.details.size + 1))
        return this
    }

    /**
     * 单步交易，一次完成交易记账过程
     * @param service
     * @param isFinished - true : 完成交易过程,扣除账户余额; false - 完成检查
     */
    fun done(service: PersonBalancePayService, isFinished: Boolean): TUserdtl {
        prepareData()
        preCheckAccount()
        return service.process(this)
    }

    /**
     * 两步交易，在 paydtl 基础完成确认过程
     * @param paydtl
     * @param status - 完成交易状态，见 TradeDict.DTL_STATUS_FAIL
     */
    fun done(paydtl: TUserdtl, status: String, service: PersonBalancePayService): TUserdtl {
        TODO("finish")
    }

    fun done(refno: String, status: String, service: PersonBalancePayService): TUserdtl {
        TODO("add process")
    }
}